# VPMXFER version 0.1

**VCPI を利用した PCI MMIO データ転送ライブラリ**

DOS 環境で PCI の MMIO にアクセスするために作ったライブラリです。
VCPIを使ってプロテクトモードに移行してデータ転送を行います。

本来の目的は PCI の MMIO とのデータのやり取りでしたが、実装上、PCI の MMIO でなくとも 10000h 以上の物理アドレスを直接指定できるメモリ空間との間であればデータ転送が可能です。

このライブラリは3条項 BSD ライセンスの下にリリースされています。


## 内容物一覧
* README.md
    * このファイルです
* LICENSE
    * ライセンスが記載されたファイルです。
* vpmxfer.asm, vpmxfer.inc, vpmxfer.h
    * 本ライブラリのソースコードです。
* vpmxfer.lib
    * ビルド済みライブラリです。
* build.bat
    * 本ライブラリをビルドするのに用いたバッチファイルです。


## VPMXFER が提供する関数

VPMXFER は次の6個の関数を提供します。呼び出し規約はすべて cdecl です。

ここでは関数の概要を纏めます。詳細はソースコードを参照してください。

* check_vcpi() 関数
    * VCPI の存在チェックを行います。
    * 引数はなく、戻り値が 0 の場合 VCPI が利用可能であることを示します。
* init_vcpi() 関数
    * 転送処理前に必要な VCPI の初期化を行います。
    * 引数はなく、初期化に成功した場合 0 が返ります。
* dispose_vcpi() 関数
    * VCPI の後始末を行うために定義した関数ですが、現状何もしていません。
* vpmxfer_4() 関数
* vpmxfer_2() 関数
* vpmxfer_1() 関数
    * データの転送を行う関数です。
    * 引数は vpmxfer.h 及び vpmxfer.inc を参照してください。
    * 名前の4, 2, 1は転送の単位となるバイト数に対応しています。
    * 引数には転送先及び転送元のアドレスと転送回数を渡します。
    * 引数のアドレスの解釈は以下のルールで行います。
        * 10000h 未満のアドレスはリニアアドレスとして扱います。
        * 10000h 以上のアドレスは物理アドレスとして扱います。
    * 10000h を跨いでの転送を行おうとした際の挙動は未定義です。


## 利用上の注意

本ライブラリには実装由来の制約が幾つかあります。
利用時は下記の点に気を付けてください。

### セグメント名とクラス名について

セグメント名とクラス名は OpenWatcom v1.90 の C コンパイラ(wcl)が生成するオブジェクトファイルとのリンクが上手くいくものにしてあります。
開発環境によっては、セグメント名とクラス名を変更して再アセンブルする必要があるかもしれません。

### 転送中の割り込みについて

現状、データ転送中は割り込みを禁止する実装になっています。
転送するデータ量が多い場合は、転送中に割り込み処理を取りこぼす可能性があるため、転送処理を複数回に分けて行う必要があります。

### 1回に転送できるデータ量の上限について

vpmxfer_*() 関数を用いて一度に転送できるサイズには上限があります。物理アドレスで指定されたメモリ空間にアクセスするために必要なページテーブルを1個しか確保していないためです。

転送先又は転送元がコンベンショナルメモリの場合は、コンベンショナルメモリ側のバッファサイズはページテーブル1個分に相当する 4MiB 以下の筈ですから、この制約は事実上無いものと考えられます。

もしも、転送先と転送元のアドレス両方とも物理アドレスを指定する場合は、指定されたメモリ空間に1個のページテーブルを通してアクセスすることになるため、転送サイズの上限は指定の物理アドレスが 4KiB 境界に乗っているという条件下で 2MiB となります。
ただし、前述のとおり転送中は割り込みを禁止していますので一度に 2MiB 転送しようとしても正しく動作しない可能性が高いと思われます。


## 利用例

以下の条件でコンベンショナルメモリから PCI デバイスの MMIO にデータ転送する場合の例を示します。

* 対象の PCI デバイスの MMIO のベースアドレスが 80000000h
* アクセスしたい MMIO 中のレジスタの位置は 68h
* 転送するデータは ds が指すセグメント中の buf という変数に格納されている
* 転送するデータの量は word サイズのデータを 8 個である

```
	; リニアアドレス計算
	xor	eax, eax
	mov	ax, ds
	shl	ax, 4
	add	ax, offset buf
	; MMIO アドレス計算
	mov	edx, 80000000h
	add	edx, 68h
	; 転送
	invoke	vpmxfer_2, edx, eax, 8
```

VPCI の存在チェックや初期化、バッファへの転送データの設定等は上記の例からは省略されています。
実際に転送を行う際はこれらの処理も適切に行う必要があります。


## 作者・更新履歴
あめさろだ(rainy-srd, r-srd), mail: r-srd@outlook.jp

* 2022-06-11, version 0.1
    * 初版

