;==============================================================================
; VPMXFER version 0.1
; vpmxfer.asm
; Copyrigt (C) 2022 あめさろだ (rainy-srd, r-srd)
; Released under the 3-clause BSD License.
;==============================================================================

;--------------------------------------------------------
; __cdecl の仕様メモ
;   * 「_」が前置される (func -> _func)
;   * 引数は右から左に push される(最後の引数が最初に push される)
;   * eax, ecx, edx は破壊されても良い

.386p

; システムテーブルレジスタ (GDT, IDT)
SYSTBLREG STRUC
	limit		dw 0
	base		dd 0
SYSTBLREG ENDS

; セグメントディスクリプタ
DESCRIPTOR STRUC
	limit_l		dw 0
	base_l		dw 0
	base_h		db 0
	acc_byte	db 0
	limit_h		db 0
	base_hh		db 0
DESCRIPTOR ENDS


ACC_CS		equ 9bh		; CS 用アクセスバイト
ACC_DS		equ 93h		; DS 用アクセスバイト
ACC_TSS386	equ 89h


DGROUP GROUP _DATA, _BSS, GDT_SEG
_DATA SEGMENT PARA USE16 PUBLIC 'DATA'
v86_cs		dd 0
v86_ds		dd 0
v86_ss		dd 0
v86_sp		dd 0

pagetbl_ptr 	dd 1000h	; ページテーブルへのポインタ (real mode)
vcpi_entry	df ?		; vcpi エントリポイント (far32 pointer)
				; [df] がオフセットで [df+4] がセレクタ

	align	4
gdtr_val	SYSTBLREG<GDT_limit, >
	align	4
idtr_val	SYSTBLREG<3ffh, 0> ; 8086 IVT

; V86 モードからプロテクトモードに移行する時に渡す引数
	align	4
argvp		label	dword
argvp_cr3	dd ?			; CR3
argvp_gdtr	dd ?			; GDTR
argvp_idtr	dd ?			; IDTR
argvp_ldtr	dw 0			; LDTR 
argvp_tr	dw offset GDT_SEG:tss_386_desc	; TR(セレクタ)
argvp_eip	dd 0			; プロテクトモードエントリ (CS:EIP)
argvp_cs	dw ?

	align	4
mapped_src_addr	dd ?
mapped_dst_addr	dd ?
xfer_cnt	dw ?
xfer_error	dw ?	; 転送の結果エラーがあれば非0
xfer_unit_size	db 0ffh	; 呼び出された vpmxfer_* に合わせて 1, 2, 4 の何れか

_DATA ENDS

; 後でリンクする場合を考慮して protect_cs, protect_ds の limit は 0ffffh
; src,dst のディスクリプタは 0～8MiB の空間を指す
GDT_SEG SEGMENT PARA USE16  PUBLIC 'DATA'
gdt_start	label byte
		DESCRIPTOR<> ; null
;gdt_alias	DESCRIPTOR<offset GDT_tail - 1, GDT_SEG, , ACC_DS>
;idt_alias	DESCRIPTOR<offset IDT_tail - 1, IDT_SEG, , ACC_DS>
tss_386_desc	DESCRIPTOR<TSS_386_limit, , , 89h>
pagetbl_alias	DESCRIPTOR<PAGETBL_limit, , , ACC_DS>
src_desc	DESCRIPTOR<07ffh, 0h, 0h, ACC_DS, 80h> ; 転送元
dst_desc	DESCRIPTOR<07ffh, 0h, 0h, ACC_DS, 80h> ; 転送先
protect_cs	DESCRIPTOR<0ffffh, _TEXT, , ACC_CS>
protect_ds	DESCRIPTOR<0ffffh, DGROUP, , ACC_DS>
protect_ss	DESCRIPTOR<0ffffh, DGROUP, , ACC_DS>
linear16M_desc	DESCRIPTOR<0fffh, 0h, , ACC_DS, 80h>
server_cs	DESCRIPTOR< , , , ACC_CS>
server_ds0	DESCRIPTOR< , , , ACC_DS>
server_ds1	DESCRIPTOR< , , , ACC_DS>
gdt_end		label byte
GDT_size	equ offset gdt_end - offset gdt_start
GDT_limit	equ offset gdt_end - offset gdt_start - 1
;GDT_tail	label byte
GDT_SEG ENDS
; セレクタ値取得用マクロ(GDT_SEG を明示したかったが度々忘れたため)
SET_SEL	MACRO reg, desc_name
	mov	reg, offset GDT_SEG:desc_name
ENDM


_BSS SEGMENT USE16 PARA 'BSS'

pagetbl_start	label byte
	org	4 * 4 * 1024
PAGETBL_limit equ 3 * 4 * 1024 - 1

	align	16
tss_start	label byte
	org	4 * 26
TSS_386_limit equ 4 * 26 - 1

_BSS ENDS


_TEXT SEGMENT USE16 PUBLIC 'CODE'
	assume cs:_TEXT, ds:DGROUP, ss:STACK

;--------------------------------------------------------
; 転送処理本体
xfer PROC NEAR C PRIVATE USES ebx edi esi es,
		dst_addr:dword, src_addr:dword, cnt:word
	; エラークリア
	mov	[xfer_error], 0
	; 転送量保存
	movzx	ecx, cnt
	mov	[xfer_cnt], cx
	; 実際に転送されるデータ量を計算
	mov	al, [xfer_unit_size]
	cmp	al, 4
	jnz	@F
	shl	ecx, 2
	jmp	calc_xfer_bytes_end
@@:	cmp	al, 2
	jnz	@F
	shl	ecx, 1
	jmp	calc_xfer_bytes_end
@@:	cmp	al, 1
	jz	calc_xfer_bytes_end
	mov	ax, 1		; 転送単位がおかしいなら return
	ret
calc_xfer_bytes_end:
	; 転送に使うリニアアドレスの計算(必要ならマッピングもやる)
	les	bx, [pagetbl_ptr]
	add	bx, 1000h		; 1番目のページテーブル(pagetbl[1])
	movzx	esi, bx			; bx 保存
	mov	edx, 400000h		; linear addr to access pagetbl[1]
	mov	eax, src_addr
	call	calc_linear_addr
	mov	[mapped_src_addr], eax
	movzx	edi, bx
	sub	edi, esi
	shl	edi, 12 - 4		; 増えたバイト数を 4KiB 単位に変換
	add	edx, edi		; bx の増加量から edx を更新
	mov	eax, dst_addr
	call	calc_linear_addr
	mov	[mapped_dst_addr], eax
	; mask NMI
	;out	50h, al
	; プロテクトモードに移行
	assume	bx:ptr DESCRIPTOR
	mov	bx, offset DGROUP:tss_386_desc
	and	[bx].acc_byte, NOT 02h ; TSS: busy -> available
	assume	bx:nothing
	mov	word ptr [v86_sp], sp
	xor	esi, esi
	mov	si, DGROUP
	shl	esi, 4
	add	esi, offset DGROUP:argvp	; argvp のリニアアドレス計算
	mov	ax, 0de0ch
	cli
	int	67h
	jmp	$
	; ここからプロテクトモード
protect_entry label far
	; セレクタ設定
	SET_SEL ax, protect_ds
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax
	SET_SEL ax, protect_ss
	mov	ss, ax
	mov	esp, [v86_sp]
	; 転送処理
	push	ds
	movzx	ecx, [xfer_cnt]
	mov	esi, [mapped_src_addr]
	mov	edi, [mapped_dst_addr]
	SET_SEL ax, src_desc
	mov	ds, ax
	SET_SEL ax, dst_desc
	mov	es, ax
	cld
	mov	al, fs:[xfer_unit_size]
	cmp	al, 1
	jne	@F
	; xfer_unit_size == 1
	db	67h	; address size override prefix
	rep	movsb
	jmp	xfer_finish
@@:	cmp	al, 2
	jne	@F
	; xfer_unit_size == 2
	db	67h
	rep	movsw
	jmp	xfer_finish
@@:	; xfer_unit_size == 4
	db	67h
	rep	movsd
xfer_finish:
	pop	ds
	; V86 モードに戻る
	mov	edx, esp
	mov	eax, [v86_ds]
	push	eax			; gs
	push	eax			; fs
	push	eax			; ds
	push	eax			; es
	push	[v86_ss]		; ss
	push	edx			; esp
	pushfd				; eflags
	push	[v86_cs]		; cs
	mov	ax, offset v86_entry
	push	eax			; eip
	SET_SEL ax, linear16M_desc
	mov	ds, ax
	SET_SEL ax, protect_ds
	mov	es, ax
	mov	ax, 0de0ch
	call	es:[vcpi_entry]		; protect_ds に置いてある
	jmp	$
	; ここから V86 モード
v86_entry label far
	sti
	; unmask nmi
	;out	52h, al
	mov	ax, [xfer_error]
	ret
xfer ENDP

;--------------------------------------------------------
; 転送処理(外部から呼び出す用)
; 引数: dst_addr,src_addr=転送元及び転送先アドレス, cnt=転送回数
;       dst_addr:dword, src_addr:dword, cnt:word
; 戻値: ax=0で成功、非0で失敗
; 仕様
; * 転送中は割り込みを禁止する
; * 10000h 未満の src と dst はリニアアドレスとして解釈される
;   * VCPI が初期化してくれたページテーブルを通して物理メモリにアクセスする
; * 10000h 以上の src と dst は物理アドレスとして解釈される
;   * 4～8MiB のページテーブルにエントリを追加して
;     4～8MiB のリニアアドレスを通して指定の物理アドレスにアクセスする
; * src の範囲及び dst の範囲が 10000h を跨ぐ場合の動作は未定義
; * 物理メモリ用のページテーブルは1個しか確保しない(4～8MiB 分)
;   * コンベンショナルメモリとの間でのデータ転送を想定しており
;     コンベンショナルメモリにバッファを取るなら 4MiB 未満なので
;     ページテーブル1個あれば足りると考えたため
;   * 1回の転送サイズ上限は、物理アドレス間での 2MiB の転送となる
;     (注: 転送開始位置が 4KiB 境界に乗っているという条件下)
;     (注: おそらく転送中に割り込みを処理できず動かないと思われる)
vpmxfer_4 PROC NEAR C PUBLIC
	mov	[xfer_unit_size], 4
	jmp	xfer
vpmxfer_4 ENDP

vpmxfer_2 PROC NEAR C PUBLIC
	mov	[xfer_unit_size], 2
	jmp	xfer
vpmxfer_2 ENDP

vpmxfer_1 PROC NEAR C PUBLIC
	mov	[xfer_unit_size], 1
	jmp	xfer
vpmxfer_1 ENDP


;--------------------------------------------------------
; VCPI 後始末
dispose_vcpi PROC NEAR C PUBLIC
	; 現状やることなし
	ret
dispose_vcpi ENDP


;--------------------------------------------------------
; VCPI 初期化
; 引数: 無し
; 戻値: ax=0 で初期化成功
init_vcpi PROC NEAR C PUBLIC USES ebx di si es
	; V86 モードのセグメントレジスタ値保存
	mov	word ptr [v86_cs], cs
	mov	word ptr [v86_ds], ds
	mov	word ptr [v86_ss], ss
	; ページテーブル領域の先頭となる 4KiB 境界を探す
	xor	eax, eax
	mov	ax, ds
	shl	eax, 4
	add	eax, offset DGROUP:pagetbl_start
	add	eax, 0fffh
	and	eax, 0fffff000h
	shr	eax, 4				; ax = segment
	mov	word ptr [pagetbl_ptr + 2], ax
	assume	bx:ptr DESCRIPTOR
	mov	bx, offset DGROUP:pagetbl_alias
	mov	[bx].base_l, ax			; 後の変換のために保存
	mov	bx, offset DGROUP:protect_ss
	mov	[bx].base_l, ss			; 後の変換のためについでに保存
	assume	bx:nothing
	; ページテーブルの領域を0埋め
	mov	es, ax
	xor	bx, bx
	mov	cx, 1000h * 3 / 4
	xor	eax, eax
@@:	mov	es:[bx], eax
	add	bx, 4
	dec	cx
	jnz	@B
	; ディスクリプタ計算(セグメント -> リニアアドレス)
	mov	dx, DGROUP
	mov	ax, offset DGROUP:gdt_start
	mov	bx, offset DGROUP:gdtr_val
	call	set_l24_to_streg
	mov	cx, GDT_size
	shr	cx, 3
	mov	bx, offset DGROUP:gdt_start
	call	conv24_cont
	; ディスクリプタ計算(TSS 用エントリの設定)
	mov	bx, offset DGROUP:tss_386_desc
	mov	dx, DGROUP
	mov	ax, offset DGROUP:tss_start
	call	set_l24_to_desc
	; VCPI プロテクトモードインターフェース取得
	les	di, [pagetbl_ptr]
	mov	si, offset DGROUP:server_cs
	mov	ax, 0de01h ; ds:si=サーバ用GDTエントリ, es:di=page_table
	int	67h
	test	ah, ah
	jnz	vcpi_get_interface_failed
	; プロテクトモード側 VCPI 呼び出しアドレス
	mov	dword ptr [vcpi_entry], ebx
	SET_SEL	ax, server_cs
	mov	word ptr [vcpi_entry + 4], ax
	; プロテクトモード移行時の引数設定(CR3)
	movzx	eax, word ptr [pagetbl_ptr + 2]	; seg of pagetbl (real mode)
	mov	es, ax
	shl	eax, 4				; linear addr of pagetbl
	mov	[argvp_cr3], eax
	; ページディレクトリエントリ設定
	add	eax, 1000h
	mov	edx, eax
	or	edx, 7				; set U/S, R/W, P bit
	mov	es:[0], edx			; add page dir entry (0-4MiB)
	add	eax, 1000h
	or	eax, 7
	mov	es:[4], eax			; add page dir entry (4-8MiB)
	; プロテクトモード移行時の引数設定(GDTR と IDTR のリニアアドレス)
	xor	eax, eax
	mov	ax, DGROUP
	shl	eax, 4
	mov	edx, eax
	add	eax, offset DGROUP:gdtr_val
	add	edx, offset DGROUP:idtr_val
	mov	[argvp_gdtr], eax
	mov	[argvp_idtr], edx
	; プロテクトモード移行時の引数設定(protect mode entry point)
	mov	[argvp_eip], offset protect_entry
	SET_SEL	[argvp_cs], protect_cs
	; 正常終了
	xor	ax, ax
init_end:
	ret
vcpi_get_interface_failed:
	mov	ax, 1
	jmp	init_end
init_vcpi ENDP


;--------------------------------------------------------
; VCPI の存在のチェック
; 引数: 無し
; 戻値: ax=0 の時 VCPI 利用可能
check_vcpi PROC NEAR C PUBLIC USES bx es
	; EMM の存在チェック
	mov	ax, 3500h + 67h
	int	21h		; es:bx = 割り込み先
	mov	bx, 0ah
	cmp	word ptr es:[bx + 0], 'ME'
	jnz	emm_not_found
	cmp	word ptr es:[bx + 2], 'XM'
	jnz	emm_not_found
	cmp	word ptr es:[bx + 4], 'XX'
	jnz	emm_not_found
	cmp	word ptr es:[bx + 6], '0X'
	jnz	emm_not_found
	; EMS が使えるか?
	mov	ah, 40h
	int	67h
	test	ah, ah
	jnz	cannot_use_ems
	; VCPI の存在チェック
	mov	ax, 0de00h
	int	67h
	test	ah, ah
	jnz	vcpi_not_found
	xor	ax, ax
check_end:
	ret
emm_not_found:
	mov	ax, 1
	jmp	check_end
cannot_use_ems:
	mov	ax, 2
	jmp	check_end
vcpi_not_found:
	mov	ax, 3
	jmp	check_end
check_vcpi ENDP


;--------------------------------------------------------
; ここから内部用関数
;--------------------------------------------------------
; 転送に使うリニアアドレスを計算する
; 引数: eax=start, ecx=xfer_bytes, es:bx=ページテーブルに追加する位置
;       edx=引数のes:[bx]を指すのに必要なリニアアドレス
; 戻値: eax=リニアアドレス
;       es:bx=未初期化のページテーブルのエントリ位置
; 破壊: 無し
calc_linear_addr PROC PRIVATE
	cmp	eax, 100000h
	jb	calc_end	; 1MiB 未満のアドレスならそのまま
	push	eax
	push	edx
	; 転送範囲を囲う 4KiB 境界を求める
	call	calc_4kib_range
	; ページテーブルにエントリを追加
	call	add_pagetbl_entry
	pop	edx
	pop	eax
	and	eax, 0fffh	; 4KiB 内でのオフセット
	add	eax, edx
calc_end:
	ret
calc_linear_addr ENDP

;--------------------------------------------------------
; 与えられたメモリ領域をマップするためのエントリをページテーブルに追加する
; 引数: eax=startの 4KiB 境界, edx=endの 4KiB 境界, es:bx=追加する位置
; 戻値: es:bx=未初期化のページテーブルのエントリ位置
; 破壊: 無し
add_pagetbl_entry PROC PRIVATE
	push	ecx
add_loop:
	cmp	eax, edx
	jae	add_end
	; 追加処理
	mov	ecx, eax
	or	ecx, 7
	mov	dword ptr es:[bx], ecx
	add	bx, 4
	add	eax, 1000h
	jmp	add_loop
add_end:
	pop	ecx
	ret
add_pagetbl_entry ENDP

;--------------------------------------------------------
; 与えられたメモリ領域を囲う範囲の 4KiB 境界を求める
; 引数: eax=base, ecx=range_len
; 戻値: eax=start, edx=end
; 破壊: 無し
calc_4kib_range PROC PRIVATE
	mov	edx, eax
	add	edx, ecx
	add	edx, 000000fffh
	and	eax, 0fffff000h
	and	edx, 0fffff000h
	ret
calc_4kib_range ENDP

;--------------------------------------------------------
; seg:off のアドレスをリニアアドレスへ変換する
; 単発(システムテーブルレジスタ用)
; 引数: ds:bx=リニアアドレスを設定するシステムテーブルレジスタのアドレス
;       dx:ax=設定したいアドレス(seg:off)
; 戻値: 無し
; 破壊: eax, edx
set_l24_to_streg PROC PRIVATE
	assume	bx:ptr SYSTBLREG
	movzx	edx, dx
	movzx	eax, ax
	shl	edx, 4
	add	edx, eax
	mov	[bx].base, edx
	assume	bx:nothing
	ret
set_l24_to_streg ENDP

; 単発(ディスクリプタのエントリ用)
; 引数: ds:bx=ディスクリプタのアドレス
;       dx:ax=設定したいアドレス(seg:off)
; 戻値: 無し
; 破壊: eax, edx
set_l24_to_desc PROC PRIVATE
	assume	bx:ptr DESCRIPTOR
	movzx	edx, dx
	movzx	eax, ax
	shl	edx, 4
	add	edx, eax
	mov	[bx].base_l, dx
	shr	edx, 16
	mov	[bx].base_h, dl
	assume	bx:nothing
	ret
set_l24_to_desc ENDP

;--------------------------------------------------------
; セグメントアドレスから 24bit リニアアドレスへの変換
; 単発
; 引数: ds:bx=addr of descriptor (base_l にセグメントアドレスを入れておく)
; 破壊: eax
conv_seg_to_l24_desc PROC PRIVATE
	assume	bx:ptr DESCRIPTOR
	movzx	eax, [bx].base_l
	shl	eax, 4
	mov	[bx].base_l, ax
	shr	eax, 16
	mov	[bx].base_h, al
	assume	bx:nothing
	ret
conv_seg_to_l24_desc ENDP

; 連続
; 引数: ds:bx=addr of descriptor table, cx=number of descriptor
; 破壊: eax, bx, cx
conv24_cont PROC PRIVATE
	assume	bx:ptr DESCRIPTOR
	test	cx, cx
	jz	conv_end
conv_loop:
	mov	al, [bx].acc_byte
	test	al, 10h		; system descriptor type ?
	jnz	do_conv24	; システムセグメントでないなら変換
	and	al, 7		; type をマスク
	cmp	al, 4		; 4以上は call/task/interrupt/trap gate
	jae	conv_loop_tail
do_conv24:
	; TSS, LDT, アクセスバイト0(これは予約だが)の場合も変換する
	call	conv_seg_to_l24_desc
conv_loop_tail:
	add	bx, 8
	loop	conv_loop
	assume	bx:nothing
conv_end:
	ret
conv24_cont ENDP

_TEXT ENDS


STACK SEGMENT USE16 PUBLIC 'STACK'
	; vpmxfer_* で push するバイト数
	org	50
STACK ENDS

END
